using System;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Components {
  // This class is effectively internal but Monocle [Tracked]
  [Tracked]
  public class PreventClimbJump : Component {
    [OnLoad]
    public static void Load() {
      On.Celeste.Player.ClimbJump += OnPlayerClimbJump;
      IL.Celeste.Player.ClimbUpdate += OnPlayerClimbUpdate;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Player.ClimbJump -= OnPlayerClimbJump;
      IL.Celeste.Player.ClimbUpdate -= OnPlayerClimbUpdate;
    }


    internal PreventClimbJump() : base(false, false) {}

    private static void OnPlayerClimbJump(On.Celeste.Player.orig_ClimbJump orig, Player self) {
      // just no-op if the component is attached
      if (self.Get<PreventClimbJump>() != null) { return; }
      orig(self);
    }

    // Because celeste is well designed there's Player.WallJumpCheck but not
    // Player.ClimbJumpCheck?
    // Just no-oping ClimbJump isn't enough because the game has a seizure when
    // trying to do it and changing player state back to normal for a frame.
    private static void OnPlayerClimbUpdate(ILContext il) {
      ILCursor cursor = new ILCursor(il);

      // I do not even want to pretend to know how precisely this works
      // This IL hook is verbatim from ExtendedVariants
      if (cursor.TryGotoNext(MoveType.After, instr => instr.MatchCallvirt<VirtualButton>("get_Pressed"))) {
        Logger.Log("bitsbolts.PreventClimbJump", $"Adding condition to kill climb jumping at {cursor.Index} in IL code for ClimbUpdate; same IL hook as ExtendedVariantMode/DisableClimbJumping");

        cursor.Emit(OpCodes.Ldarg_0);
        // load moveX which is private
        cursor.Emit(OpCodes.Ldarg_0);
        cursor.Emit(OpCodes.Ldfld, typeof(Player).GetField("moveX", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Instance));

        // inject a method to mod the result of the jump button check
        cursor.EmitDelegate<Func<bool, Player, int, bool>>(PreventClimbJump.OnJumpButtonCheck);
      } else {
        Logger.Log(LogLevel.Error, "bitsbolts.PreventClimbJump", $"IL@{cursor.Index}: Hook failed to find VirtualButton Pressed getter callvirt in Player.ClimbUpdate");
      }
    }

    private static bool OnJumpButtonCheck(bool origVal, Player self, int moveX) {
      if (self.Get<PreventClimbJump>() != null) {
        if (moveX == 0 - self.Facing) {
          // This will lead to a wall jump. We want to kill climb jumping. So let it go
          return origVal;
        }

        // let the game believe Jump is not pressed, so it won't return the player to the Normal state (leading to a weird animation / sound effect).
        return false;
      }

      // nothing to do
      return origVal;
    }
  }
}
