using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [Tracked(true)]
  public class CreditsScroller : Entity {
    public readonly Vector2 Size;
    public float Scale = 1f;

    public float Scroll = 0f;
    // public float EndScroll = 800f;
    public float AutoMaxSpeed = 75f;
    public float ManualMaxSpeed = 75f * 6f;
    public bool ManualScrolling = true;
    public bool CanConfirm = false;
    private Tween ConfirmFadeTween;
    private float scrollManualTimeout = 0f;
    private float scrollSpeed = 0f;

    private VirtualRenderTarget target = null;

    public CreditsScroller() : this(
      new Vector2(Screen.Width * .25f, Screen.Height * .1f), new Vector2(Screen.Width * .5f, Screen.Height * .8f)
    ) {}

    public CreditsScroller(Vector2 pos, Vector2 size) {
      base.Position = pos;
      this.Size = size;

      base.AddTag(TagsExt.SubHUD | Tags.Global | Tags.TransitionUpdate | Tags.FrozenUpdate);
      base.Add(new BeforeRenderHook(this.BeforeRender));
    }

    public new float Width { get => this.Size.X; }
    public new float Height { get => this.Size.Y; }

    // public Hitbox ComponentsHitbox {
    //   get {
    //     Vector2 min = new Vector2(Single.Infinity, Single.Infinity);
    //     Vector2 max = new Vector2(Single.NegativeInfinity, Single.NegativeInfinity);
    //     foreach (var graphic in this.Components.GetAll<GraphicsComponent>()) {
    //       // if (graphic is BoundedGraphicsComponent) {
    //       //   var hitbox = graphic.Hitbox;
    //       //   min.X = hitbox.Top
    //       // }
    //     }
    //     return new Hitbox()
    //   }
    // }

    public float EndScroll {
      get {
        float maxY = Single.NegativeInfinity;
        foreach (var graphic in this.Components.GetAll<GraphicsComponent>()) {
          if (graphic is BoundedGraphicsComponent bounded) {
            maxY = Math.Max(bounded.Position.Y + bounded.Height, maxY);
          } else if (graphic is Image image) {
            maxY = Math.Max(image.Position.Y + image.Height, maxY);
          } else {
            Logger.Log(LogLevel.Error, "bitsbolts.CreditsScroller", $"Unknown graphics component subclass attached to entity!");
          }
        }
        return maxY + this.Height / 2;
      }
    }


    public override void Removed(Scene scene) {
      if (target != null) {
        this.target.Dispose();
        target = null;
      }
      base.Removed(scene);
    }


    public override void Update() {
      base.Update();

      this.Scroll += this.scrollSpeed * Engine.DeltaTime;
      this.Scroll = this.Scroll.Clamp(0f, this.EndScroll);

      var accel = 2.5f;

      if (this.scrollManualTimeout > 0f) {
        this.scrollManualTimeout = Math.Max(this.scrollManualTimeout - Engine.DeltaTime, 0f);
      } else {
        this.scrollSpeed = Calc.Approach(this.scrollSpeed, this.AutoMaxSpeed, this.AutoMaxSpeed * accel * Engine.DeltaTime);
      }

      if (this.ManualScrolling) {
        if (Input.MenuDown.Check) {
          this.scrollManualTimeout = 3f;
          this.scrollSpeed = Calc.Approach(this.scrollSpeed, this.ManualMaxSpeed, this.ManualMaxSpeed * accel * Engine.DeltaTime);
        } else if (Input.MenuUp.Check) {
          this.scrollManualTimeout = 3f;
          this.scrollSpeed = Calc.Approach(this.scrollSpeed, -this.ManualMaxSpeed, this.ManualMaxSpeed * accel * Engine.DeltaTime);
        } else if (this.scrollManualTimeout > 0f) {
          this.scrollSpeed = Calc.Approach(this.scrollSpeed, 0f, this.ManualMaxSpeed * accel * Engine.DeltaTime);
        }
      }

      if (!this.CanConfirm && this.Scroll >= this.EndScroll) {
        this.CanConfirm = true;
        base.Add(
          this.ConfirmFadeTween = Tween.Create(Tween.TweenMode.Persist, Ease.CubeIn, 0.5f, true)
        );
      }

      if (this.CanConfirm && Input.MenuConfirm.Pressed) {
        // todo: maybe find a better sxf to play
        Audio.Play("event:/ui/main/button_lowkey");
        // todo: call callback ig?
      }
    }

    public void BeforeRender() {
      if (this.target == null) {
        this.target = VirtualContent.CreateRenderTarget(
          name: "bitsbolts_credits",
          width: (int) this.Width,
          height: (int) this.Height,
          depth: false,
          preserve: true,
          multiSampleCount: 0
        );
      }

      Vector2 offset = new Vector2(0, this.target.Height - this.Scroll);

      Engine.Graphics.GraphicsDevice.SetRenderTarget(this.target);
      Engine.Graphics.GraphicsDevice.Clear(Color.Transparent);

      Draw.SpriteBatch.Begin(
        SpriteSortMode.Deferred,  // sortMode
        BlendState.AlphaBlend,  // blendState
        null,  // samplerState
        null,  // depthStencilState
        null,  // rasterizerState
        null,  // effect
        Matrix.CreateTranslation(new Vector3(offset, 0f))  // transformMatrix
      );
      DynamicData.For(this.Components).Invoke("Render");
      Draw.SpriteBatch.End();
    }

    public override void Render() {
      // do not call base.Render; we render all components in BeforeRender

      if (this.target != null) {
        Draw.SpriteBatch.Draw(
          texture: this.target,
          position: base.Position,
          sourceRectangle: new Rectangle?(this.target.Bounds),
          color: Color.White,
          // we don't care to support rotation for this
          rotation: 0f,
          origin: Vector2.Zero,
          scale: 1f,
          effects: SpriteEffects.None,
          layerDepth: 0f
        );
      }

      if (this.CanConfirm) {
        Input.GuiButton(Input.MenuConfirm, Input.PrefixMode.Latest).DrawCentered(
          base.Position + this.Size * 0.90f - new Vector2(0, (this.ConfirmFadeTween.Eased - 1) * 20f),
          Color.White * this.ConfirmFadeTween.Eased
        );
      }
    }
  }
}
