using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/GrablessBerry")]
  [RegisterStrawberry(tracked: false, blocksCollection: true)]
  [Tracked]
  public class GrablessBerry : PersistentBerry {
    [OnLoad]
    public static new void Load() {
      Everest.Events.Level.OnCreatePauseMenuButtons += OnLevelOnCreatePauseMenuButtons;
    }

    [OnUnload]
    public static new void Unload() {
      Everest.Events.Level.OnCreatePauseMenuButtons -= OnLevelOnCreatePauseMenuButtons;
    }


    public static new ParticleType P_Glow;
    public static new ParticleType P_GhostGlow;

    public readonly bool PreventWallGrab = true;
    public readonly bool PreventWallJump = false;
    public readonly bool PreventClimbJump = false;

    private PreventWallGrab PreventWallGrabComponent = null;
    private PreventWallJump PreventWallJumpComponent = null;
    private PreventClimbJump PreventClimbJumpComponent = null;

    internal class SessionData {
      public bool PreventWallGrab, PreventWallJump, PreventClimbJump;
      public bool OnlySpawnLevelComplete, OnlySpawnNoDeaths;
      public Vector2 BerryStart;
    }

    public GrablessBerry(EntityData data, Vector2 offset, EntityID id) : this(data.Position + offset, id, false) {
      this.PreventWallGrab = data.Bool("preventWallGrab", true);
      this.PreventWallJump = data.Bool("preventWallJump", false);
      this.PreventClimbJump = data.Bool("preventClimbJump", false);

      base.OnlySpawnLevelComplete = data.Bool("onlySpawnLevelComplete", true);
      base.OnlySpawnNoDeaths = data.Bool("onlySpawnNoDeaths", true);

      // tmp migration
      if (data.Bool("preventWallKick", false)) {
        this.PreventWallJump = true;
        this.PreventClimbJump = true;
      }
    }

    [RestoreEntity("Bitsbolts/GrablessBerry", typeof(SessionData))]
    internal GrablessBerry(EntityID id, SessionData data, string name)
    : this(data.BerryStart, id, true) {
      this.PreventWallGrab = data.PreventWallGrab;
      this.PreventWallJump = data.PreventWallJump;
      this.PreventClimbJump = data.PreventClimbJump;

      base.OnlySpawnLevelComplete = data.OnlySpawnLevelComplete;
      base.OnlySpawnNoDeaths = data.OnlySpawnNoDeaths;
    }

    internal GrablessBerry(Vector2 start, EntityID eid, bool restored)
    : base(start, eid, restored) {
      base.SpritePath = "bitsbolts_grabless_berry";
      base.GhostSpritePath = "bitsbolts_ghost_grabless_berry";

      if (P_Glow == null) {
        P_Glow = new ParticleType(Strawberry.P_Glow) {
          Color = Calc.HexToColor("5F4980"),
          Color2 = Calc.HexToColor("85A1A6")
        };
        P_GhostGlow = new ParticleType(Strawberry.P_GhostGlow) {
          Color = Calc.HexToColor("818E9E"),
          Color2 = Calc.HexToColor("645974")
        };
      }
      base.Glow = P_Glow;
      base.GhostGlow = P_GhostGlow;
    }

    public override SerializedEntity<object> Serialize() {
      return new SerializedEntity<object> {
        Name = "Bitsbolts/GrablessBerry",
        Data = (object) new SessionData {
          PreventWallGrab = this.PreventWallGrab,
          PreventWallJump = this.PreventWallJump,
          PreventClimbJump = this.PreventClimbJump,
          OnlySpawnLevelComplete = base.OnlySpawnLevelComplete,
          OnlySpawnNoDeaths = base.OnlySpawnNoDeaths,
          BerryStart = base.BerryStart,
        }
      };
    }

    public override void OnGainLeader(Entity leader) {
      base.OnGainLeader(leader);
      if (leader is Player player) {
        if (this.PreventWallGrab) {
          player.Add(PreventWallGrabComponent = new PreventWallGrab());
        }
        if (this.PreventWallJump) {
          player.Add(PreventWallJumpComponent = new PreventWallJump());
        }
        if (this.PreventClimbJump) {
          player.Add(PreventClimbJumpComponent = new PreventClimbJump());
        }
      }
    }

    public override void OnLoseLeader(Entity leader) {
      base.OnLoseLeader(leader);
      if (this.PreventWallGrabComponent != null) {
        this.PreventWallGrabComponent.RemoveSelf();
        PreventWallGrabComponent = null;
      }
      if (this.PreventWallJumpComponent != null) {
        this.PreventWallJumpComponent.RemoveSelf();
        PreventWallJumpComponent = null;
      }
      if (this.PreventClimbJumpComponent != null) {
        this.PreventClimbJumpComponent.RemoveSelf();
        PreventClimbJumpComponent = null;
      }
    }

    public string BerryName {
      get {
        if (this.PreventWallGrab && this.PreventWallJump) {
          return "bitsbolts_grablessberry_grab_wall".DialogOrKey();
        } else if (this.PreventWallGrab) {
          return "bitsbolts_grablessberry_grab".DialogOrKey();
        } else if (this.PreventWallJump && this.PreventClimbJump) {
          return "bitsbolts_grablessberry_wall".DialogOrKey();
        } else if (this.PreventClimbJump) {
          return "bitsbolts_grablessberry_climb".DialogOrKey();
        } else {
          return "bitsbolts_grablessberry_wall".DialogOrKey();
        }
      }
    }

    private static void OnLevelOnCreatePauseMenuButtons(Level level, TextMenu menu, bool minimal) {
      if (minimal) { return; }

      foreach (GrablessBerry berry in level.Tracker.GetEntities<GrablessBerry>()) {
        if (berry == null || !berry.Follower.HasLeader) { continue; }

        TextMenu.Button item = new TextMenu.Button(
          "bitsbolts_grablessberry_stop".DialogRawOrKey().FormatWith(berry.BerryName)
        ) {
          OnPressed = () => {
            level.Paused = false;
            level.PauseMainMenuOpen = false;
            menu.RemoveSelf();

            DynamicData.For(berry as Strawberry).Set("start", berry.BerryStart);
            berry.Follower.Leader.LoseFollower(berry.Follower);
          }
        };
        menu.Add(item);
      }
    }


    [Command("bb_giveglb", "[bits & bolts] Gives the player a grabless berry")]
    private static void OnCommandGive() {
      if (Engine.Scene is Level level) {
        Player player = level.Tracker.GetEntity<Player>();
        if (player != null) {
          var data = new EntityData();
          data.Position = player.Position + new Vector2(0f, -16f);
          data.ID = Calc.Random.Next();
          data.Name = "bitsbolts/GrablessBerry";
          data.Values = new Dictionary<string, object>();
          data.Values["_bitsbolts_persistentberry_restored"] = true;

          var berry = new GrablessBerry(
            data, Vector2.Zero, new EntityID(level.Session.Level, data.ID)
          );
          level.Add(berry);
        }
      } else {
        Engine.Commands.Log("not in a map dumbass", Color.Yellow);
      }
    }


    [Command("bb_giveglbjump", "[bits & bolts] Gives the player a grabless berry with no walljumping")]
    private static void OnCommandGiveNoWallKick() {
      if (Engine.Scene is Level level) {
        Player player = level.Tracker.GetEntity<Player>();
        if (player != null) {
          var data = new EntityData();
          data.Position = player.Position + new Vector2(0f, -16f);
          data.ID = Calc.Random.Next();
          data.Name = "bitsbolts/GrablessBerry";
          data.Values = new Dictionary<string, object>();
          data.Values["_bitsbolts_persistentberry_restored"] = true;
          data.Values["preventWallGrab"] = true;
          data.Values["preventWallJump"] = true;
          data.Values["preventClimbJump"] = true;

          var berry = new GrablessBerry(
            data, Vector2.Zero, new EntityID(level.Session.Level, data.ID)
          );
          level.Add(berry);
        }
      } else {
        Engine.Commands.Log("not in a map dumbass", Color.Yellow);
      }
    }
  }
}
