using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  // A key which largely acts like the Temporary Key from Frost Helper, but
  // doesn't respawn after use, is intended to be used on normal key doors, and
  // survives screen transitions. It disappears from the player and respawns at
  // the key's spawnpoint when the player dies. If it was used to unlock a door,
  // it will never respawn again for the current session.
  [Tracked]
  [CustomEntity("Bitsbolts/MostlyTemporaryKey")]
  public class MostlyTemporaryKey : Key {
    private static ILHook ILMapEditorRenderKeysHook;

    [OnLoad]
    public static void Load() {
      On.Celeste.Key.RegisterUsed += OnKeyRegisterUsed;

      var RenderKeys = typeof(Celeste.Editor.MapEditor).GetMethod("RenderKeys", BindingFlags.NonPublic | BindingFlags.Instance);
      if (RenderKeys == null) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey", $"Looks like the monomod police finally caught up to me :(");
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey", $"tears up my everest warranty... you'll never catch me alive!");
      } else {
        ILMapEditorRenderKeysHook = new ILHook(RenderKeys, ILMapEditorRenderKeys);
      }
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Key.RegisterUsed -= OnKeyRegisterUsed;

      ILMapEditorRenderKeysHook?.Dispose();
    }


    public struct SessionData { public Vector2 Start; }

    public Sprite Sprite;
    public Vector2 Start;
    private bool restored = false;

    public MostlyTemporaryKey(EntityData data, Vector2 offset, EntityID id)
    : this(data.Position + offset, data.NodesOffset(offset), id) {}

    [RestoreEntity("Bitsbolts/MostlyTemporaryKey", typeof(SessionData), loaderOnly: true)]
    internal MostlyTemporaryKey(EntityID id, SessionData data, string name)
    : this(data.Start, new Vector2[0], id) {
      this.restored = true;
    }

    private MostlyTemporaryKey(Vector2 start, Vector2[] nodes, EntityID id)
    : base(start, id, nodes) {
      this.Start = start;

      var dd = DynamicData.For(this as Key);

      var sprite = dd.Get<Sprite>("sprite");
      base.Remove(sprite);
      base.Add(this.Sprite = GFX.SpriteBank.Create("bitsbolts_mostly_temporary_key"));
      dd.Set("sprite", this.Sprite);

      var follower = base.Get<Follower>();
      follower.OnGainLeader = (Action) Delegate.Combine(
        follower.OnGainLeader, new Action(this.OnGainLeader)
      );
      follower.OnLoseLeader = (Action) Delegate.Combine(
        follower.OnLoseLeader, new Action(this.OnLoseLeader)
      );

      // ghetto hook the Key PlayerCollider
      PlayerCollider collider = base.Get<PlayerCollider>();
      Action<Player> OnCollideOrig = collider.OnCollide;
      collider.OnCollide = player => {
        // undo registering the key in Keys and DoNotLoad to prevent persistence
        OnCollideOrig(player);

        (base.Scene as Level).Session.Keys.Remove(base.ID);
        // donotload will be readded in ongain leader ig?
        (base.Scene as Level).Session.DoNotLoad.Remove(base.ID);
      };
    }

    public override void Awake(Scene scene) {
      if (this.restored) {
        Player player = scene.Tracker.GetEntity<Player>();
        if (player == null) {
          // wtf...?
          base.RemoveSelf();
          return;
        }

        this.Position = player.Position;
        base.Depth = Depths.Top;
        this.Collidable = false;
        player.Leader.GainFollower(base.Get<Follower>());
      }
    }


    private static void OnKeyRegisterUsed(On.Celeste.Key.orig_RegisterUsed orig, Celeste.Key self) {
      if (self is MostlyTemporaryKey key) {
        orig(self);
        key.RegisterUsed();
      } else {
        orig(self);
      }
    }

    public new void RegisterUsed() {
      base.Collidable = false;
      (base.Scene as Level).Session.DoNotLoad.Add(this.ID);
      MainModule.Session.StoredEntities.Remove(base.ID);
    }


    public void OnGainLeader() {
      (base.Scene as Level).Session.Keys.Remove(base.ID);
      (base.Scene as Level).Session.DoNotLoad.Add(base.ID);
      MainModule.Session.StoredEntities.Set(base.ID, new SerializedEntity<SessionData> {
        Name = "Bitsbolts/MostlyTemporaryKey",
        Data = new SessionData { Start = this.Start }
      });
    }


    public void OnLoseLeader() {
      // we can't kill the key now...
      if (base.IsUsed || base.Turning) { return; }

      (base.Scene as Level).Session.DoNotLoad.Remove(base.ID);
      MainModule.Session.StoredEntities.Remove(base.ID);

      var follower = base.Get<Follower>();
      if (follower.Leader != null) {
        follower.Leader.LoseFollower(follower);
      }

      Add(new Coroutine(RespawnRoutine(), true));
    }

    // gfx from frost helper my beloved
    public IEnumerator RespawnRoutine() {
      var level = base.Scene as Level;

      base.Collidable = false;
      Audio.Play("event:/game/general/seed_poof", Position);
      this.Sprite.Scale = Vector2.One * .5f;

      yield return .05f;

      Input.Rumble(RumbleStrength.Medium, RumbleLength.Medium);

      for (int i = 0; i < 6; i++) {
        float dir = Calc.Random.NextFloat((float) Math.PI * 2f);
        level.ParticlesFG.Emit(
          StrawberrySeed.P_Burst,
          1,
          Position + Calc.AngleToVector(dir, 4f),
          Vector2.Zero,
          dir
        );
      }
      this.Sprite.Scale = Vector2.Zero;
      base.Visible = false;

      yield return .3f;

      Audio.Play("event:/game/general/seed_reappear", Position);
      Position = this.Start;
      this.Sprite.Scale = Vector2.One;
      base.Visible = true;
      base.Collidable = true;
      level.Displacement.AddBurst(Position, 0.2f, 8f, 28f, 0.2f, null, null);
    }

    private static void ILMapEditorRenderKeys(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.Before, instr => instr.MatchLdfld<LevelData>("Bounds"))) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey", $"IL@{cursor.Index}: Hook failed to find find ldfld LevelData.Bounds in MapEditor.RenderKeys!");
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey", $"The monomod police must be hot on my trail...");
        return;
      }

      cursor.Emit(OpCodes.Dup);  // LevelData
      cursor.Emit(OpCodes.Ldarg_0);  // MapEditor
      cursor.EmitDelegate<Action<LevelData, Celeste.Editor.MapEditor>>(
        (LevelData levelData, Celeste.Editor.MapEditor editor) => {
          var bounds = levelData.Bounds;
          var basePos = new Vector2(bounds.X, bounds.Y);
          var keys = DynamicData.For(editor).Get<List<Vector2>>("keys");

          if (keys == null) {
            Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey.ILMapEditorRenderKeys_delegate", $"Failed to find find MapEditor.RenderKeys.Keys!");
            Logger.Log(LogLevel.Error, "Bitsbolts.Entities.MostlyTemporaryKey", $"The monomod police must have swiped the keys away from my bare hands...");
            return;
          }

          keys.AddRange(
            levelData.Entities
              .Where(edata => edata.Name == "Bitsbolts/MostlyTemporaryKey")
              .Select(edata => (basePos + edata.Position) / 8)
          );
        }
      );
    }
  }
}
