using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [Tracked(true)]
  public abstract class PersistentBerry : Strawberry, IStrawberry {
    [OnLoad]
    public static void Load() {
      On.Celeste.Strawberry.OnCollect += OnStrawberryOnCollect;

      // Any mod that fucks with this function and prevents player death will
      // fuck up our berry's golden status and may cause a golden berry restart.
      using (new DetourContext { Before = { "*" } }) {
        On.Celeste.Player.Die += OnPlayerDie;
      }
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Strawberry.OnCollect -= OnStrawberryOnCollect;
      On.Celeste.Player.Die -= OnPlayerDie;
    }

    public ParticleType Glow;
    public ParticleType GhostGlow;
    public string SpritePath;
    public string GhostSpritePath;
    public bool Restored;
    public bool OnlySpawnLevelComplete = true;
    public bool OnlySpawnNoDeaths = false;
    public readonly EntityID Eid;
    public readonly Vector2 BerryStart;

    private Entity prevLeader = null;

    private static readonly Vector2 restoredOffset = new Vector2(8, -16);


    public PersistentBerry(Vector2 position, EntityID eid, bool restored)
    : base(
      new EntityData {
        Position = position,
        ID = eid.ID,
        Name = "Bitsbolts/PersistentBerry",
        Nodes = new Vector2[0],
        Values = new Dictionary<string, object>(),
      },
      Vector2.Zero,
      eid
    ) {
      this.Eid = eid;
      this.Restored = restored;
      this.BerryStart = position;

      base.Follower.PersistentFollow = true;
      DynamicData.For(this as Strawberry).Set("Golden", true);

      // ghetto hook the Strawberry PlayerCollider
      PlayerCollider collider = base.Get<PlayerCollider>();
      var OnCollideOrig = collider.OnCollide;
      collider.OnCollide = (player) => {
        // the berry shouldn't flip the "grabbed golden" switch in session; from cu2
        bool prevGrabbedGolden = (base.Scene as Level).Session.GrabbedGolden;
        OnCollideOrig(player);
        (base.Scene as Level).Session.GrabbedGolden = prevGrabbedGolden;
      };

      base.Follower.OnGainLeader = (Action) Delegate.Combine(
        base.Follower.OnGainLeader,
        new Action(() => {
          if (base.Follower.Leader?.Entity is Player player) {
            this.OnGainLeader(player);
          }
        })
      );

      base.Follower.OnLoseLeader = (Action) Delegate.Combine(
        base.Follower.OnLoseLeader,
        new Action(() => { this.OnLoseLeader(prevLeader); })
      );
    }


    public bool Collected {
      get => DynamicData.For(this as Strawberry).Get<bool>("collected");
      set => DynamicData.For(this as Strawberry).Set("collected", value);
    }


    public override void Added(Scene scene) {
      var level = scene as Level;
      var sess = (scene as Level).Session;
      var save = SaveData.Instance;

      if (this.Restored) {
        var spawn = level.Session.RespawnPoint ?? level.GetSpawnPoint(
          new Vector2(level.Bounds.Left, level.Bounds.Top)
        );

        // place the berry at the player's current spawn right above the head
        var start = spawn + PersistentBerry.restoredOffset;
        DynamicData.For(this as Strawberry).Set("start", start);
        base.Position = start;

        // in case it wasn't added to the donotload list; eg. via commands
        sess.DoNotLoad.Add(this.Eid);
      }


      // From cu2: remove the berry under the vanilla golden conditions
      if (
        !this.Restored && (
          // died in a further screen
          (this.OnlySpawnLevelComplete && sess.FurthestSeenLevel != sess.Level && sess.Deaths != 0)
          // hasn't completed the level first yet
          || (
            this.OnlySpawnNoDeaths && !save.CheatMode
            && !save.Areas_Safe[sess.Area.ID].Modes[(int) sess.Area.Mode].Completed
          )
        )
      ) {
        RemoveSelf();
      }

      base.Added(scene);

      Sprite sprite;
      if (!SaveData.Instance.CheckStrawberry(this.Eid)) {
        sprite = GFX.SpriteBank.Create(SpritePath);
      } else {
        sprite = GFX.SpriteBank.Create(GhostSpritePath);
        sprite.Color = Color.White * 0.8f;
      }

      Sprite oldSprite = DynamicData.For(this as Strawberry).Get<Sprite>("sprite");
      // continue vanilla code
      sprite.OnFrameChange = oldSprite.OnFrameChange;

      // hotswap the sprites
      base.Remove(oldSprite);
      DynamicData.For(this as Strawberry).Set("sprite", sprite);
      (this as Entity).Add(sprite);
      sprite.Play("idle");
    }


    public override void Awake(Scene scene) {
      base.Awake(scene);

      // manually add the berry to the player; this prevents the collect sound
      // and some motion from playing after death
      if (this.Restored) {
        var player = (scene as Level).Tracker.GetEntity<Player>();
        player.Leader.GainFollower(this.Follower);
        base.Position = player.Position + PersistentBerry.restoredOffset;

        DynamicData.For(this as Strawberry).Get<Wiggler>("wiggler").Start();
        base.Depth = Depths.Top;
      }
    }


    public override void Removed(Scene scene) {
      // if (prevLeader != null) { this.OnLoseLeader(prevLeader); }
      base.Removed(scene);
    }


    public override void Update() {
      var origGoldGlow = Strawberry.P_GoldGlow;
      var origGhostGlow = Strawberry.P_GhostGlow;
      Strawberry.P_GoldGlow = this.Glow;
      Strawberry.P_GhostGlow = this.GhostGlow;
      base.Update();
      Strawberry.P_GoldGlow = origGoldGlow;
      Strawberry.P_GhostGlow = origGhostGlow;
    }

    public abstract SerializedEntity<object> Serialize();

    public virtual void OnGainLeader(Entity leader) {
      prevLeader = leader;

      // this will be called again when restoring the berry, but since these are
      // hashsets, we really don't give a shit
      (this.Scene as Level).Session.DoNotLoad.Add(this.Eid);
      MainModule.Session.StoredEntities[this.Eid] = this.Serialize();
      this.AddTag(Tags.Global);
      this.Collidable = false;
    }


    public virtual void OnLoseLeader(Entity leader) {
      this.prevLeader = null;
      this.Collidable = true;

      // This event is called on death ig... it fucks up persistence so uhhhh
      if (leader is Player player && player.Dead) { return; }
      // OnCollect has special logic which handles this case
      if (this.Collected) { return; }

      (this.Scene as Level).Session.DoNotLoad.Remove(this.Eid);
      MainModule.Session.StoredEntities.Remove(this.Eid);

      this.RemoveTag(Tags.Global);


      // temporarily disable collider so the player doesn't pick it again
      this.Collidable = false;
      Alarm.Set(this, 0.6f, () => {
        this.Collidable = true;
      });
    }


    public void QueueCollect() {
      base.Add(new Coroutine(this.CollectRoutine()));
    }

    private IEnumerator<float> CollectRoutine() {
      if (!this.Follower.HasLeader || this.Collected) { yield break; }

      while (this.Follower.DelayTimer > 0f) {
        yield return this.Follower.DelayTimer;
      }

      // ig the berry was already collected? OnCollect will likely fail so bail
      if (!this.Follower.HasLeader || this.Collected) { yield break; }

      yield return 0.15f;
      (this as Strawberry).OnCollect();

      // just make extra sure it's marked
      (this.Scene as Level).Session.DoNotLoad.Add(this.Eid);
      MainModule.Session.StoredEntities.Remove(this.Eid);
      yield break;
    }


    private static void OnStrawberryOnCollect(On.Celeste.Strawberry.orig_OnCollect orig, Strawberry self) {
      // Strawberry.OnCollect is not marked as override
      if (self is PersistentBerry berry) {
        berry.OnCollect();
        orig(berry);
      } else {
        orig(self);
      }
    }

    public virtual new void OnCollect() {
      this.Collected = false;
      // just make extra sure it's marked
      (this.Scene as Level).Session.DoNotLoad.Add(this.Eid);
      MainModule.Session.StoredEntities.Remove(this.Eid);
      this.RemoveTag(Tags.Global);
    }


    private static PlayerDeadBody OnPlayerDie(
      On.Celeste.Player.orig_Die orig, Player self, Vector2 direction,
      bool evenIfInvincible, bool registerDeathInStats
    ) {
      var followers = new List<Follower>(self.Leader.Followers);

      // we need to swap out the golden state for orig to prevent it from
      // restarting the chapter
      foreach (var follower in followers) {
        if (follower.Entity is PersistentBerry berry) {
          DynamicData.For(berry).Set("Golden", false);
        }
      }

      var body = orig(self, direction, evenIfInvincible, registerDeathInStats);

      foreach (var follower in followers) {
        if (follower.Entity is PersistentBerry berry) {
          DynamicData.For(berry).Set("Golden", true);
        }
      }

      // if the player didn't die than don't fuck with the berry position
      if (body == null) { return null; }

      // the player is now removed from the scene, so self.Leader.Followers is
      // now invalid, so we unfourtunantly need to cache it to use it after orig

      foreach (var follower in followers) {
        if (follower.Entity is PersistentBerry berry) {
          var level = berry.Scene as Level;
          var spawn = level.Session.RespawnPoint ?? level.GetSpawnPoint(
            new Vector2(level.Bounds.Left, level.Bounds.Top)
          );

          // place the berry at the player's current spawn right above the head
          var start = spawn + PersistentBerry.restoredOffset;
          DynamicData.For(berry as Strawberry).Set("start", start);
          // the berry will approach start after losing the player leader
        }
      }

      return body;
    }
  }
}
