using System;
using System.Collections;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Flagstrs = System.Collections.Generic.Dictionary<string, bool>;
using Bitsbolts.Components;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/SaveDataFlagSync")]
  [Tracked]
  public class SaveDataFlagSync : Entity {
    [OnLoad]
    public static void Load() {
      On.Celeste.SaveData.BeforeSave += OnSaveDataBeforeSave;
      Everest.Events.Level.OnExit += OnLevelExit;
      On.Celeste.LevelEnter.Go += OnLevelEnterGo;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.SaveData.BeforeSave -= OnSaveDataBeforeSave;
      Everest.Events.Level.OnExit -= OnLevelExit;
    }


    public readonly Flagstrs Flags;
    public readonly bool LoadFlags = true;
    public readonly bool SaveFlags = true;
    public readonly bool Raw = false;

    public SaveDataFlagSync(EntityData data, Vector2 offset) : base() {
      this.Flags = data.Attr("flags", "").ParseFlagstrs(allowInversions: false);
      this.LoadFlags = data.Bool("load", true);
      this.SaveFlags = data.Bool("save", true);
      this.Raw = data.Bool("raw", false);

      this.Collidable = false;
      this.Visible = false;
    }

    // do flag load in added to allow entities that may rely on these flags in
    // awake the ability to use them early
    public override void Added(Scene scene) {
      base.Added(scene);

      if (this.LoadFlags) {
        foreach (var flag in this.Flags.Keys) {
          if (MainModule.Session.SaveDataFlagsLoaded.Contains(flag)) { continue; }

          if (this.Raw) {
            (scene as Level).Session.SetFlag(flag, SaveData.Instance.Flags.Contains(flag));
          } else {
            var flag_prefixed = $"{(scene as Level).Session.Area.SID}:{flag}";
            (scene as Level).Session.SetFlag(flag, SaveData.Instance.Flags.Contains(flag_prefixed));
          }
        }

        MainModule.Session.SaveDataFlagsLoaded.UnionWith(this.Flags.Keys);
      }

      if (this.SaveFlags) {
        (
          this.Raw ? MainModule.Session.SaveDataFlagsToSaveRaw
            : MainModule.Session.SaveDataFlagsToSave
        ).UnionWith(this.Flags.Keys);
      }
    }

    private static void UpdateSaveData(Session session, SaveData savedata) {
      foreach (var flag in MainModule.Session.SaveDataFlagsToSave) {
        var flag_prefixed = $"{session.Area.SID}:{flag}";
        if (session.GetFlag(flag)) {
          savedata.Flags.Add(flag_prefixed);
        } else {
          savedata.Flags.Remove(flag_prefixed);
        }
      }

      foreach (var flag in MainModule.Session.SaveDataFlagsToSaveRaw) {
        if (session.GetFlag(flag)) {
          savedata.Flags.Add(flag);
        } else {
          savedata.Flags.Remove(flag);
        }
      }
    }

    private static void OnSaveDataBeforeSave(On.Celeste.SaveData.orig_BeforeSave orig, SaveData save) {
      orig(save);

      // We should never ever cause a crash in this hook.
      // Crashing here means the game won't save the player's level progress.

      // Saving may happen outside of a level
      if (save.CurrentSession_Safe == null || MainModule.Session == null) { return; }

      // In the case of CollabUtils2, it somehow fucking manages to call
      // BeforeSave when the our ModuleSession is reset, even though vanilla never
      // does it. Rely on OnLevelExit to handle saving properly, which should
      // have ran before this. At this point, UpdateSaveData should be noop.
      SaveDataFlagSync.UpdateSaveData(save.CurrentSession_Safe, save);
    }

    private static void OnLevelExit(
      Level level, LevelExit exit, LevelExit.Mode mode,
      Session session, HiresSnow snow
    ) {
      // Similarly, we should also never crash here.

      if (session == null || MainModule.Session == null) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.SaveDataFlagSync", $"wtf session is null? Why the fuck is LevelExit being called with a null session? Not saving flags!");
        return;
      }

      if (SaveData.Instance == null) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.SaveDataFlagSync", $"wtf savedata is null? Why the fuck is LevelExit being called with null savedata? Not saving flags!");
        return;
      }

      SaveDataFlagSync.UpdateSaveData(session, SaveData.Instance);

      // the level ended, so steamroll all of our flag tracking the data out!
      // it won't matter past this point, however if a mod has the wise idea to
      // reuse our session without serializing & deserializing it, then these
      // ephemeral arrays won't get reset. so do it ourselves just in case...
      MainModule.Session.SaveDataFlagsLoaded.Clear();
      MainModule.Session.SaveDataFlagsToSave.Clear();
      MainModule.Session.SaveDataFlagsToSaveRaw.Clear();
    }

    private static void OnLevelEnterGo(
      On.Celeste.LevelEnter.orig_Go orig, Session session, bool fromSaveData
    ) {
      // Catch the case when someone is loading a level without calling LevelExit
      // or without calling Everest.Events.Level.OnExit
      //
      // This is the absolute last chance we have to save flags before the old
      // Scene gets wiped out.
      if (
        Engine.Scene != null && Engine.Scene is Level oldLevel
        && MainModule.Session != null
        && SaveData.Instance.CurrentSession_Safe != null
        && SaveData.Instance.CurrentSession_Safe == oldLevel.Session
      ) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.SaveDataFlagSync", $"Another mod is doing something nasty and didn't call LevelExit or the OnExit event when switching levels. Saving savedata flags anyways, but fix your shit!");

        UpdateSaveData(oldLevel.Session, SaveData.Instance);
      }

      orig(session, fromSaveData);
    }
  }
}
