using System;
using System.Linq;
using System.Collections.Generic;
using System.Reflection;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using YamlDotNet.Serialization;


namespace Bitsbolts {
  [AttributeUsage(AttributeTargets.Constructor, AllowMultiple = false, Inherited = false)]
  public class RestoreEntity : Attribute {
    public string Name;
    public Type DataType;
    public bool LoaderOnly;
    public RestoreEntity(string name, Type dataType, bool loaderOnly = true) {
      this.Name = name;
      this.DataType = dataType;
      this.LoaderOnly = loaderOnly;
    }
  }

  // todo: consider dispatching this on EverestModule.SerializeSession or smth
  // [AttributeUsage(AttributeTargets.Method, AllowMultiple = false, Inherited = false)]
  // public class OnStoreEntities : Attribute {}


  public static class SessionEntity {
    internal delegate Entity RestoreCtor(EntityID id, object data, string name);
    internal struct Restorer {
      internal RestoreCtor Func;
      internal bool LoaderOnly;
      internal Type DataType;
    }
    internal static Dictionary<string, Restorer> Restorers = new Dictionary<string, Restorer>();


    [OnLoad]
    public static void Load() {
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
    }

    [OnInitialize]
    public static void Initialize() {
      foreach (var type in typeof(Attributes).Assembly.GetTypesSafe()) {
        foreach (var method in type.GetMethods(BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.Public)) {
          foreach (RestoreEntity attr in method.GetCustomAttributes(typeof(RestoreEntity), false)) {
            SessionEntity.Restorers[attr.Name] = new SessionEntity.Restorer {
              Func = (RestoreCtor) Delegate.CreateDelegate(typeof(RestoreCtor), method),
              LoaderOnly = attr.LoaderOnly,
              DataType = attr.DataType,
            };

            Logger.Log("bitsbolts.SessionEntity", $"Found entity restore attribute {attr.Name} => {method.FullTypeName()}");
          }
        }

        foreach (var ctor in type.GetConstructors(BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic)) {
          foreach (RestoreEntity attr in ctor.GetCustomAttributes(typeof(RestoreEntity), false)) {
            SessionEntity.Restorers[attr.Name] = new SessionEntity.Restorer {
              Func = (RestoreCtor) ((a, b, c) => (Entity) ctor.Invoke(new object[] { a, b, c })),
              LoaderOnly = attr.LoaderOnly,
              DataType = attr.DataType,
            };

            Logger.Log("bitsbolts.SessionEntity", $"Found entity restore attribute {attr.Name} => {ctor.FullTypeName()}");
          }
        }
      }
    }


    public static Entity Restore(
      EntityID id, SerializedEntity<object> data, Level level, bool fromLoader = false
    ) {
      Entity entity;
      if (SessionEntity.Restorers.ContainsKey(data.Name)) {
        var restorer = SessionEntity.Restorers[data.Name];
        if (restorer.LoaderOnly && !fromLoader) { return null; }

        // todo: cast data to SerializedEntity<data.Type>
        entity = (Entity) restorer.Func.DynamicInvoke(id, data.Data, data.Name);
      } else {
        throw new KeyNotFoundException($"Entity restorer for data {data.Name} is not registered!").Carded();
      }
      level.Add(entity);
      return entity;
    }


    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level,
      Player.IntroTypes playerIntro, bool isFromLoader
    ) {
      var fromLoader = playerIntro != Player.IntroTypes.Transition && isFromLoader;

      // add all ids to donotload so they don't get loaded by LoadLevel
      level.Session.DoNotLoad.UnionWith(MainModule.Session.StoredEntities.Keys);

      orig(level, playerIntro, isFromLoader);

      foreach (var data in MainModule.Session.StoredEntities) {
        SessionEntity.Restore(data.Key, data.Value, level, fromLoader);
      }
    }
  }
}
