using System;
using System.Reflection;
using System.Text.RegularExpressions;

using Microsoft.Xna.Framework;
using Monocle;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts {
  public static class Util {
    internal static string DialogOrKey(this string key, Language lang = null) {
      return key.DialogCleanOrNull(lang) ?? key;
    }

    internal static string DialogRawOrKey(this string key, Language lang = null) {
      return Dialog.Has(key, lang) ? Dialog.Get(key, lang) : key;
    }

    internal static string DialogNlOrKey(this string key, Language lang = null) {
      return key.DialogRawOrKey().Replace("{break}", "{n}");
    }

    internal static string DialogFormat(this string key, params object[] args) {
      return Regex.Replace(
        key.DialogRawOrKey(),
        // I'm so sorry
        // \{([^\d\{\}]+|(?:[^\{\}]*[^\d\{\}]+\d[^\{\}]*)|(?:[^\{\}]*\d[^\d\{\}]+[^\{\}]*))\}
        "\\{([^\\d\\{\\}]+|(?:[^\\{\\}]*[^\\d\\{\\}]+\\d[^\\{\\}]*)|(?:[^\\{\\}]*\\d[^\\d\\{\\}]+[^\\{\\}]*))\\}",
        "{{$1}}"
      ).FormatWith(args);
    }

    internal static string DialogNlFormat(this string key, params object[] args) {
      return key.DialogFormat(args).Replace("{break}", "{n}");
    }


    internal static string FormatWith(this string format, params object[] args) {
      try {
        return String.Format(format, args);
      } catch (FormatException) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Util.FormatWith", $"Formatting string {format} failed with args: {string.Join(", ", args)}");
        return format;
      }
    }

    internal static T Clamp<T>(this T val, T min, T max) where T : IComparable<T> {
      if (val.CompareTo(min) < 0) {
        return min;
      } else if (val.CompareTo(max) > 0) {
        return max;
      } else {
        return val;
      }
    }

    internal static bool IsCore() {
      return (
        Assembly.GetEntryAssembly()?
        .GetCustomAttribute<System.Runtime.Versioning.TargetFrameworkAttribute>()?
        .FrameworkName
      ).StartsWith(".NETCore");
    }

    internal static Color HexToColor(this string hex) {
      hex = hex.TrimStart('#');
      if (hex.Length < 6) {
        throw new IndexOutOfRangeException("Hex colors must contain at least 6 characters.").Carded();
      }

      float r = (float)(Calc.HexToByte(hex[0]) * 16 + Calc.HexToByte(hex[1])) / 255f;
      float g = (float)(Calc.HexToByte(hex[2]) * 16 + Calc.HexToByte(hex[3])) / 255f;
      float b = (float)(Calc.HexToByte(hex[4]) * 16 + Calc.HexToByte(hex[5])) / 255f;

      if (hex.Length < 8) {
        return new Color(r, g, b);
      }

      float a = (float)(Calc.HexToByte(hex[6]) * 16 + Calc.HexToByte(hex[7])) / 255f;
      // premultiply the alpha
      return new Color(r, g, b) * a;
    }

    internal static string ToHexAlpha(this Color color) {
      return $"#{color.R:x2}{color.G:x2}{color.B:x2}{color.A:x2}";
    }

    internal static Hitbox HitboxFromCorners(Vector2 vec1, Vector2 vec2) {
      float p1x = Math.Min(vec1.X, vec2.X);
      float p1y = Math.Min(vec1.Y, vec2.Y);
      float p2x = Math.Max(vec1.X, vec2.X);
      float p2y = Math.Max(vec1.Y, vec2.Y);

      return new Hitbox(p2x - p1x, p2y - p1y, p1x, p1y);
    }

    internal static (float, float, float, float) RectFromCorners(Vector2 vec1, Vector2 vec2) {
      float p1x = Math.Min(vec1.X, vec2.X);
      float p1y = Math.Min(vec1.Y, vec2.Y);
      float p2x = Math.Max(vec1.X, vec2.X);
      float p2y = Math.Max(vec1.Y, vec2.Y);

      return (p1x, p1y, p2x - p1x, p2y - p1y);
    }

    internal static Rectangle RectangleFromCorners(Vector2 vec1, Vector2 vec2) {
      (var p1, var p2, var w, var h) = RectFromCorners(vec1, vec2);
      return new Rectangle((int) p1, (int) p2, (int) w, (int) h);
    }

    internal static Vector2 Perpendicular(this Vector2 vec) {
      return Vector2.Normalize(new Vector2(vec.Y, -vec.X));
    }

    internal static Vector2 Project(Vector2 target, Vector2 projectee) {
      var target_norm = Vector2.Normalize(target);
      return Vector2.Multiply(target_norm, Vector2.Dot(projectee, target_norm));
    }
  }
}
