using System;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;
using static Bitsbolts.Triggers.ResizeTriggersMode;


namespace Bitsbolts.Triggers {
  [CustomEntity("Bitsbolts/ResizeTriggers")]
  public class ResizeTriggers : Trigger {
    public readonly ResizeTriggersMode Mode;
    private Hitbox resizeTarget;

    public ResizeTriggers(EntityData data, Vector2 offset) : base(data, offset) {
      switch (data.Attr("mode", "screen")) {
        case "screen": this.Mode = CoversScreen; break;
        case "this": this.Mode = CoversThis; break;
        case "area": this.Mode = CoversArea; break;
        case "map": this.Mode = CoversMap; break;
      }

      if (this.Mode == CoversArea && data.Nodes.Length != 2) {
        throw new IndexOutOfRangeException("Bitsbolts/ResizeTriggers must have exactly 2 nodes when mode is covers area").Carded();
      }

      if (this.Mode == CoversArea) {
        this.resizeTarget = Util.HitboxFromCorners(data.Nodes[0] + offset, data.Nodes[1] + offset);
      }

      // ModCompat workaround for TriggerTrigger compatibility.
      base.Add(new AwakePriority(-10));
    }

    // Running the resize before other Entity's Awake gives better compatibility
    // with triggers, handling the case where triggers might do position
    // calculations or entity tracking in Awake.
    public override void Awake(Scene scene) {
      var level = scene as Level;
      switch (this.Mode) {
        case CoversScreen: resizeTarget = new Hitbox(
          level.Bounds.Width,
          level.Bounds.Height,
          level.Bounds.X,
          level.Bounds.Y
        ); break;
        case CoversThis: resizeTarget = new Hitbox(
          base.Collider.Width,
          base.Collider.Height,
          base.Position.X,
          base.Position.Y
        ); break;
        case CoversArea: break;
        case CoversMap: resizeTarget = new Hitbox(
          level.Session.MapData.Bounds.Width,
          level.Session.MapData.Bounds.Height,
          level.Session.MapData.Bounds.X,
          level.Session.MapData.Bounds.Y
        );
        break;
      }

      if (this.Mode == CoversScreen || this.Mode == CoversMap) {
        resizeTarget.Position.Y -= 24f;  // account for the raised ceiling
        resizeTarget.Height += 24f + 12f;  // and for the lower death plane
      }

      base.Collidable = true;
      foreach (Trigger trigger in scene.Tracker.GetEntities<Trigger>()) {
        // don't resize ourselves dumbass
        if (trigger is ResizeTriggers) { continue; }

        var dd = DynamicData.For(trigger);

        // Ensure that we can target this trigger with CollideCheck
        var origCollidable = trigger.Collidable;
        trigger.Collidable = true;

        // Check if the trigger was already resized by a previous ResizeTriggers
        // This can happen if ResizeTriggers overlap or if CoversScreen is used
        bool resized = (bool) (
          // Dynamicdata has a temper tantrum when it fails to get a value type,
          // so use a boxed boolean type instead
          dd.Get<bool?>("_bitsbolts_resizedtrigger")
          ?? false
        );

        if (!resized && base.CollideCheck(trigger)) {
          dd.Set("_bitsbolts_resizedtrigger", (bool?) true);

          // hope it's not using a weird collider that we need to preserve :(
          trigger.Collider = new Hitbox(
            resizeTarget.Width, resizeTarget.Height, 0, 0
          );
          trigger.Position.X = resizeTarget.Position.X;
          trigger.Position.Y = resizeTarget.Position.Y;
        }

        trigger.Collidable = origCollidable;
      }

      base.Collidable = false;
      base.RemoveSelf();
    }
  }

  public enum ResizeTriggersMode {
    CoversScreen,
    CoversThis,
    CoversArea,
    CoversMap,
  }
}
