using System;
using System.Linq;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;
using Flagstrs = System.Collections.Generic.Dictionary<string, bool>;


namespace Bitsbolts.Triggers {
  [CustomEntity("Bitsbolts/SkipFlingBird")]
  [Tracked]
  public class SkipFlingBird : Trigger {
    [OnLoad]
    public static void Load() {
      On.Monocle.Tracker.Initialize += TrackerInitialize;
    }

    [OnUnload]
    public static void Unload() {
      On.Monocle.Tracker.Initialize -= TrackerInitialize;
    }

    private static void TrackerInitialize(On.Monocle.Tracker.orig_Initialize orig) {
      orig();
      MonocleExt.TrackEntity<FlingBird>(inherited: true);
    }


    public readonly int Index;
    public readonly bool Instant;
    public readonly Vector2 FindPos;
    private FlingBird target = null;

    public SkipFlingBird(EntityData data, Vector2 offset) : base(data, offset) {
      this.Index = data.Int("index");
      this.Instant = data.Bool("instant", false);

      this.FindPos = data.Nodes.Length > 0 ? data.NodesOffset(offset)[0] : this.Position;

      var flags = data.Attr("flags", "").ParseFlagstrs();
      if (flags.Count > 0) { base.Add(new FlagstrTrigger(flags)); }
    }

    public override void Awake(Scene scene) {
      base.Awake(scene);
      this.target = scene.Tracker.GetNearestEntity<FlingBird>(this.FindPos);
      if (this.target == null) {
        Logger.Log(LogLevel.Warn, "bitsbolts.SkipFlingBird", $"Fling bird entity not found in level! Removing self.");
        base.RemoveSelf();
      }
    }

    public override void OnEnter(Player player) {
      base.OnEnter(player);

      if (this.target != null && this.target.Scene == base.Scene) {
        var dd = DynamicData.For(this.target);
        var curIndex = dd.Get<int>("segmentIndex");
        var state = dd.Get<BirdStates>("state");
        var nodes = this.target.NodeSegments;

        if (this.Index < 0 || this.Index >= nodes.Count) {
          Logger.Log(LogLevel.Error, "bitsbolts.SkipFlingBird", $"Provided index {this.Index} is out of bounds! Entity max index is {nodes.Count}!");
          return;
        }

        if (
          state != BirdStates.Wait
          || this.Index == curIndex  // don't bother; it won't do shit anyways
          || (this.Index < curIndex)  // don't go backwards; it is janky
          || this.Index == 0  // Index 0 crashes MoveRoutine, and should never happen
        ) { return; }

        if (this.Instant) {
          dd.Set("segmentIndex", this.Index);
          target.Position = nodes[this.Index - 1][nodes[this.Index - 1].Length - 1];
        } else {
          // make Skip think we are at the node just before the desired node
          // it uses entity position in the tween instead of the last node
          dd.Set("segmentIndex", this.Index - 1);
          dd.Invoke("Skip");
        }
      }
    }

    private enum BirdStates {
      Wait,
      Fling,
      Move,
      WaitForLightningClear,
      Leaving
    }
  }
}
